import logging
import sys

FORMATTER = logging.Formatter("%(asctime)s — %(levelname)s — %(message)s")

def get_logger(name):
  """
  return the logger

  :param name: (str) name of a logger
  :return:
  """
  logger = logging.getLogger(name)

  # if (logger.hasHandlers()):
  #   logger.handlers.clear()

  handler = logging.StreamHandler(sys.stdout)
  handler.setFormatter(FORMATTER)

  logger.addHandler(handler)
  logger.setLevel(logging.DEBUG)
  logger.propagate = False

  return logger



from typing import List, Type

import torch
import torch.nn as nn

def mlp(
  sizes: List[int],
  activation: Type[nn.Module] = nn.Tanh,
  output_activation: Type[nn.Module] = nn.Identity
) -> nn.Module:
  layers: List[nn.Module] = []

  for j in range(len(sizes)-1):
    current_activation = activation if j < len(sizes)-2 else output_activation
    layers += [nn.Linear(sizes[j], sizes[j+1]), current_activation()]

  return nn.Sequential(*layers)



import torch
import torch.nn as nn
from torch.distributions.categorical import Categorical

class Policy(nn.Module):
  """
  A policy.

  The policy network selects action based on the observation of the environment.
  It uses a PyTorch neural network module to fit the function of pi(s).

  :param network: (nn.Module) The network.
  :param optimizer: (torch.optim.Optimizer) The optimizer.
  """

  def __init__(
    self,
    network: nn.Module,
    optimizer: torch.optim.Optimizer
  ):
    super().__init__()

    self.network = network
    self.optimizer = optimizer

  def forward(self, observation: torch.Tensor) -> Categorical:
    """
    Forward pass in policy

    :param observation: (torch.Tensor) The current observation of the environment
    """
    logits: torch.Tensor = self.network(observation)
    distribution: Categorical = Categorical(logits=logits)

    return distribution

  def predict(self, observation: torch.Tensor) -> torch.Tensor:
    """
    Selects action based on the observation of the environment.

    :param observation: (torch.Tensor) The current observation of the environment
    """
    with torch.no_grad():
      logits: torch.Tensor = self.network(observation)
    distribution: Categorical = Categorical(logits=logits)

    action: torch.Tensor = distribution.sample()

    return action



from typing import List, Tuple, Type

import torch
import torch.nn as nn
import gym

from rl_replicas.common.torch_net import mlp

class ValueFunction(nn.Module):
  """
  A Value Function

  :param optimizer: (torch.optim.Optimizer) The optimizer.
  :param network: (nn.Module) The network.
  """

  def __init__(
    self,
    network: nn.Module,
    optimizer: torch.optim.Optimizer
  ):
    super().__init__()

    self.network = network
    self.optimizer = optimizer

  def forward(self, observation: torch.Tensor) -> torch.Tensor:
    """
    Forward pass in value function

    :param observation: (torch.Tensor) The current observation of the environment
    """
    value: torch.Tensor = self.network(observation)

    return value



from abc import ABC, abstractmethod
import os
import time
from typing import Any, Optional, List, Tuple
from typing_extensions import TypedDict

import gym
import numpy as np
import torch
from torch.distributions.categorical import Categorical
from torch.utils.tensorboard import SummaryWriter

from rl_replicas import log
from rl_replicas.common.policy import Policy
from rl_replicas.common.utils import seed_random_generators
from rl_replicas.common.value_function import ValueFunction
from rl_replicas.common.utils import discount_cumulative_sum, seed_random_generators, gae

logger = log.get_logger(__name__)

class OneEpochExperience(TypedDict):
  observations: torch.Tensor
  actions: torch.Tensor
  advantages: torch.Tensor
  discounted_returns: torch.Tensor
  episode_returns: List[float]
  episode_lengths: List[int]

class OnPolicyAlgorithm(ABC):
  """
  The base of RL algorithms

  :param policy: (Policy) The policy
  :param value_function: (ValueFunction) The value function
  :param env: (gym.Env or str) The environment to learn from
  :param gamma: (float) Discount factor
  :param gae_lambda: (float) Factor for trade-off of bias vs variance for Generalized Advantage Estimator. Equivalent to classic advantage when set to 1.
  :param seed: (int) The seed for the pseudo-random generators
  :param n_value_gradients (int): Number of gradient descent steps to take on value function per epoch.
  """
  def __init__(
    self,
    policy: Policy,
    value_function: ValueFunction,
    env: gym.Env,
    gamma: float,
    gae_lambda: float,
    seed: Optional[int],
    n_value_gradients: int
  ) -> None:
    self.policy = policy
    self.value_function = value_function
    self.env = env
    self.gamma = gamma
    self.gae_lambda = gae_lambda
    if seed is not None:
      self.seed: int = seed
    self.n_value_gradients = n_value_gradients

    if seed is not None:
      self._seed()

  def _seed(self) -> None:
    seed_random_generators(self.seed)
    self.env.action_space.seed(self.seed)
    self.env.seed(self.seed)

  def learn(
    self,
    epochs: int = 50,
    steps_per_epoch: int = 4000,
    output_dir: str = '.',
    tensorboard: bool = False,
    model_saving: bool = False
  ) -> None:
    """
    Learn the model

    :param epochs: (int) The number of epochs (equivalent to number of policy updates) to perform
    :param steps_per_epoch: (int) The number of steps to run per epoch; in other words, batch size is steps.
    :param output_dir: (str) The directory of output
    :param tensorboard: (bool) Whether or not to log for tensorboard
    :param model_saving: (bool) Whether or not to save trained model (Save and overwrite at each end of epoch)
    """
    self.writer: SummaryWriter = None
    if tensorboard:
      logger.info('Set up tensorboard')
      os.makedirs(output_dir, exist_ok=True)
      tensorboard_path: str = os.path.join(output_dir, 'tensorboard')
      self.writer = SummaryWriter(tensorboard_path)

    start_time: float = time.time()
    self.current_total_steps: int = 0
    self.current_total_episodes: int = 0

    for current_epoch in range(epochs):

      one_epoch_experience: OneEpochExperience = self.collect_experience_one_epoch(steps_per_epoch)

      if model_saving:
        logger.info('Set up model saving')
        os.makedirs(output_dir, exist_ok=True)
        model_path: str = os.path.join(output_dir, 'model.pt')

        logger.info('Save model')
        torch.save({
            'epoch': current_epoch+1,
            'total_steps': self.current_total_steps,
            'policy_state_dict': self.policy.network.state_dict(),
            'policy_optimizer_state_dict': self.policy.optimizer.state_dict(),
            'value_fn_state_dict': self.value_function.network.state_dict(),
            'value_fn_optimizer_state_dict': self.value_function.optimizer.state_dict()
          },
          model_path)

      episode_returns: List[float] = one_epoch_experience['episode_returns']
      episode_lengths: List[int] = one_epoch_experience['episode_lengths']

      logger.info('Epoch: {}'.format(current_epoch+1))

      logger.info('Total env interactions: {:<8.3g}'.format(self.current_total_steps))
      logger.info('Total episodes:         {:<8.3g}'.format(self.current_total_episodes))

      logger.info('Average Episode Return: {:<8.3g}'.format(np.mean(episode_returns)))
      logger.info('Episode Return STD:     {:<8.3g}'.format(np.std(episode_returns)))
      logger.info('Max Episode Return:     {:<8.3g}'.format(np.max(episode_returns)))
      logger.info('Min Episode Return:     {:<8.3g}'.format(np.min(episode_returns)))

      logger.info('Average Episode Length: {:<8.3g}'.format(np.mean(episode_lengths)))

      self.train(one_epoch_experience)

      logger.info('Time:                   {:<8.3g}'.format(time.time()-start_time))

    if tensorboard:
      self.writer.flush()
      self.writer.close()

  def collect_experience_one_epoch(self, steps_per_epoch: int) -> OneEpochExperience:
    one_epoch_experience: OneEpochExperience = {
      'observations': None,
      'actions': None,
      'advantages': None,
      'discounted_returns': None,
      'episode_returns': [],
      'episode_lengths': []
    }

    observations_list: List[torch.Tensor] = []
    actions_list: List[torch.Tensor] = []

    advantages_ndarray: np.ndarray = np.zeros(steps_per_epoch, dtype=np.float32)
    discounted_returns_ndarray: np.ndarray = np.zeros(steps_per_epoch, dtype=np.float32)

    episode_returns_list: List[float] = []
    episode_lengths_list: List[int] = []

    # Variables on the current episode
    rewards: List[float] = []
    values: List[float] = []
    episode_length: int = 0

    observation: np.ndarray = self.env.reset()

    for current_step in range(steps_per_epoch):
      observation_tensor: torch.Tensor = torch.from_numpy(observation).float()

      observations_list.append(observation_tensor)

      with torch.no_grad():
        policy_dist: Categorical = self.policy(observation_tensor)
        value: torch.Tensor = self.value_function(observation_tensor)

      values.append(value.detach().item())

      action: torch.Tensor = policy_dist.sample()

      actions_list.append(action)

      action_ndarray = action.detach().numpy()
      reward: float
      episode_done: bool
      observation, reward, episode_done, _ = self.env.step(action_ndarray)

      rewards.append(reward)

      episode_length += 1

      self.current_total_steps += 1

      epoch_ended: bool = current_step == steps_per_epoch-1

      if episode_done or epoch_ended:
        if epoch_ended and not(episode_done):
          logger.warn('The trajectory cut off at {} steps on the current episode'.format(episode_length))

        last_value_float: float
        if epoch_ended:
          observation_tensor = torch.from_numpy(observation).float()

          with torch.no_grad():
            last_value: torch.Tensor = self.value_function(observation_tensor)

          last_value_float = last_value.detach().item()
        else:
          last_value_float = 0.0
        values.append(last_value_float)
        rewards.append(last_value_float)

        episode_slice = slice(current_step-episode_length+1, current_step+1)

        # Calculate advantage over an episode
        values_ndarray: np.ndarray = np.asarray(values)
        rewards_ndarray: np.ndarray = np.asarray(rewards)
        episode_advantage: np.ndarray = gae(rewards_ndarray, self.gamma, values_ndarray, self.gae_lambda)

        advantages_ndarray[episode_slice] = episode_advantage

        # Calculate rewards-to-go over an episode, to be targets for the value function
        episode_discounted_return: np.ndarray = discount_cumulative_sum(rewards, self.gamma)[:-1]
        discounted_returns_ndarray[episode_slice] = episode_discounted_return

        episode_true_return: float = np.sum(rewards).item()

        episode_returns_list.append(episode_true_return)
        episode_lengths_list.append(episode_length)

        if episode_done and self.writer:
          self.writer.add_scalar(
            'env/episode_true_return',
            episode_true_return,
            self.current_total_steps
          )
          self.writer.add_scalar(
            'env/episode_length',
            episode_length,
            self.current_total_steps
          )

        if episode_done:
          self.current_total_episodes += 1

        observation, episode_length = self.env.reset(), 0
        rewards, values = [], []

    one_epoch_experience['observations'] = torch.stack(observations_list)
    one_epoch_experience['actions'] = torch.stack(actions_list)
    one_epoch_experience['advantages'] = torch.from_numpy(advantages_ndarray)
    one_epoch_experience['discounted_returns'] = torch.from_numpy(discounted_returns_ndarray)
    one_epoch_experience['episode_returns'] = episode_returns_list
    one_epoch_experience['episode_lengths'] = episode_lengths_list

    return one_epoch_experience

  @abstractmethod
  def train(
    self,
    one_epoch_experience: OneEpochExperience
  ) -> None:
    """
    Consume experience on the current epoch and update train algorithm.

    Implemented by individual algorithms.

    :param one_epoch_experience: (OneEpochExperience) Collected experience on one epoch.
    """
    raise NotImplementedError

  def predict(
    self,
    observation: np.ndarray
  ) -> np.ndarray:
    """
    Get the action(s) from an observation which are sampled under the current policy.

    :param observation: the input observation
    :return: the model's action
    """
    observation_tensor: torch.Tensor = torch.from_numpy(observation).float()
    action: torch.Tensor = self.policy.predict(observation_tensor)
    action_ndarray: np.ndarray = action.detach().numpy()

    return action_ndarray



from typing import Callable, Iterable, List, Tuple

import torch
from torch.optim import Optimizer
import numpy as np

from rl_replicas import log
from rl_replicas.common.utils import unflatten_tensors

logger = log.get_logger(__name__)

class ConjugateGradientOptimizer(Optimizer):
  """
  Performs constrained optimization via backtracking line search.

  The search direction is computed using a conjugate gradient algorithm,
  which gives x = H^{-1}g, where H is a second order approximation of the
  constraint and g is the gradient of the loss function.

  :param params: (Iterable) Iterable of parameters to optimize.
  :param max_constraint: (float) Maximum constraint value.
  :param n_conjugate_gradients: (int) The number of conjugate gradient iterations used to calculate H^-1 g
  :param max_backtracks: (int) Max number of iterations for backtrack linesearch.
  :param backtrack_ratio: (float) backtrack ratio for backtracking line search.
  :param hvp_damping_coefficient: (float) Artifact for numerical stability, should be smallish. Adjusts Hessian-vector product calculation: H -> H + hvp_damping_coefficient*I.
  """
  def __init__(
    self,
    params: Iterable[torch.Tensor],
    max_constraint: float = 0.01,
    n_conjugate_gradients: int = 10,
    max_backtracks: int = 15,
    backtrack_ratio: float = 0.8,
    hvp_damping_coefficient: float = 1e-5
  ):
    # no default values
    defaults: dict = {}
    super().__init__(params, defaults)
    self.max_constraint = max_constraint
    self.n_conjugate_gradients = n_conjugate_gradients
    self.max_backtracks = max_backtracks
    self.backtrack_ratio = backtrack_ratio
    self.hvp_damping_coefficient = hvp_damping_coefficient

  def step( # type: ignore[override]
    self,
    loss_function: Callable,
    kl_divergence_function: Callable
  ):
    """
    Performs a single optimization step.

    :param loss_function: (Callable) Function to compute the loss.
    :param kl_divergence_function: (Callable) Function to compute the kl divergence.
    """
    # Collect trainable parameters and gradients
    params: List[torch.Tensor] = []
    grads: List[torch.Tensor] = []
    for group in self.param_groups:
      for p in group['params']:
        if p.grad is not None:
          params.append(p)
          grads.append(p.grad.reshape(-1))
    flat_loss_grads: torch.Tensor = torch.cat(grads)

    # Build Hessian-vector-product function
    hessian_vector_product_function = self._build_hessian_vector_product(kl_divergence_function, params)

    # Compute step direction
    step_direction = self._conjugate_gradient(hessian_vector_product_function, flat_loss_grads)

    # Replace nan with 0.
    step_direction[step_direction.ne(step_direction)] = 0.

    # Compute step size
    step_size = np.sqrt(2.0 * self.max_constraint *
                        (1. / (torch.dot(step_direction, hessian_vector_product_function(step_direction)) + 1e-8)))

    if np.isnan(step_size):
      step_size = 1.

    descent_step = step_size * step_direction

    # Update parameters using backtracking line search
    self._backtracking_line_search(params, descent_step, loss_function, kl_divergence_function)

  @property # type: ignore[override]
  def state(self):
    """
    dict: The hyper-parameters of the optimizer.
    """
    return {
      'max_constraint': self.max_constraint,
      'n_conjugate_gradients': self.n_conjugate_gradients,
      'max_backtracks': self.max_backtracks,
      'backtrack_ratio': self.backtrack_ratio,
      'hvp_damping_coefficient': self.hvp_damping_coefficient,
    }

  @state.setter
  def state(self, state): # type: ignore[override]
    self.max_constraint = state.get('max_constraint', 0.01)
    self.n_conjugate_gradients = state.get('n_conjugate_gradients', 10)
    self.max_backtracks = state.get('max_backtracks', 15)
    self.backtrack_ratio = state.get('backtrack_ratio', 0.8)
    self.hvp_damping_coefficient = state.get('hvp_damping_coefficient', 1e-5)

  def __setstate__(self, state: dict):
    """
    Restore the optimizer state.

    :param state: (dict) State dictionary.
    """
    if 'hvp_damping_coefficient' not in state['state']:
      logger.warn('Resuming ConjugateGradientOptimizer with lost state.')
    self.defaults = state['defaults']
    # Set the fields manually so that the setter gets called.
    self.state = state['state']
    self.param_groups = state['param_groups']

  def _build_hessian_vector_product(
    self,
    hessian_target_vector_function: Callable,
    params: List[torch.Tensor]
  ) -> Callable:
    param_shapes: List[torch.Size] = [p.shape or torch.Size([1]) for p in params]
    hessian_target_vector = hessian_target_vector_function()
    hessian_target_vector_grads: Tuple[torch.Tensor, ...] = torch.autograd.grad(hessian_target_vector, params, create_graph=True)

    def _eval(vector):
      """
      The evaluation function.

      :param vector (torch.Tensor): The vector to be multiplied with Hessian.
      :return: (torch.Tensor) The product of Hessian of function f and v.
      """
      unflatten_vector: torch.Tensor = unflatten_tensors(vector, param_shapes)

      assert len(hessian_target_vector_grads) == len(unflatten_vector)
      grad_vector_product_list: List[torch.Tensor] = []
      for g, x in zip(hessian_target_vector_grads, unflatten_vector):
        single_grad_vector_product = torch.sum(g * x)
        grad_vector_product_list.append(single_grad_vector_product)

      grad_vector_product = torch.sum(torch.stack(grad_vector_product_list))

      hvp: List[torch.Tensor] = list(torch.autograd.grad(grad_vector_product, params, retain_graph=True))
      for i, (hx, p) in enumerate(zip(hvp, params)):
        if hx is None:
          hvp[i] = torch.zeros_like(p)

      flat_output: torch.Tensor = torch.cat([h.reshape(-1) for h in hvp])
      return flat_output + self.hvp_damping_coefficient * vector

    return _eval

  def _conjugate_gradient(
    self,
    hessian_vector_product_function: Callable,
    b: torch.Tensor,
    residual_tol: float = 1e-10
  ):
    """
    Use Conjugate Gradient iteration to solve Ax = b. Demmel p 312.

    :param hessian_vector_product_function: (Callable) A function to compute Hessian vector product.
    :param b: (torch.Tensor) Right hand side of the equation to solve.
    :param residual_tol: (float) Tolerence for convergence.

    :return: (torch.Tensor) Solution x* for equation Ax = b.
    """
    x = torch.zeros_like(b)

    r = b.clone()
    p = b.clone()
    rdotr = torch.dot(r, r)

    for _ in range(self.n_conjugate_gradients):
      z = hessian_vector_product_function(p)
      v = rdotr / torch.dot(p, z)
      x += v * p
      r -= v * z
      newrdotr = torch.dot(r, r)
      mu = newrdotr / rdotr
      p = r + mu * p

      rdotr = newrdotr
      if rdotr < residual_tol:
        break
    return x

  def _backtracking_line_search(
    self,
    params: List[torch.Tensor],
    descent_step: float,
    loss_function: Callable,
    constraint_function: Callable
  ):
    previous_params: List[torch.Tensor] = [p.clone() for p in params]
    ratio_list: np.ndarray = self.backtrack_ratio ** np.arange(self.max_backtracks)
    loss_before: torch.Tensor = loss_function()

    param_shapes: List[torch.Size] = [p.shape or torch.Size([1]) for p in params]
    descent_step_list: List[torch.Tensor] = unflatten_tensors(descent_step, param_shapes)
    assert len(descent_step_list) == len(params)

    for ratio in ratio_list:
      for step, previous_param, param in zip(descent_step_list, previous_params, params):
        step = ratio * step
        new_param = previous_param.data - step
        param.data = new_param.data

      new_loss = loss_function()
      constraint = constraint_function()

      if (new_loss < loss_before and constraint <= self.max_constraint):
        break

    if (torch.isnan(new_loss)
        or torch.isnan(constraint)
        or new_loss >= loss_before
        or constraint >= self.max_constraint):
      logger.warn('Line search condition violated. Rejecting the step!')
      if torch.isnan(new_loss):
        logger.warn('Violated because loss is NaN')
      if torch.isnan(constraint):
        logger.warn('Violated because constraint is NaN')
      if new_loss >= loss_before:
        logger.warn('Violated because loss not improving')
      if constraint >= self.max_constraint:
        logger.warn('Violated because constraint is violated')

      for previous_param, param in zip(previous_params, params):
        param.data = previous_param.data



import copy
from typing import Optional

import torch
from torch.nn import functional as F
from torch.distributions.categorical import Categorical
import gym

from rl_replicas.common.base_algorithms.on_policy_algorithm import OnPolicyAlgorithm, OneEpochExperience
from rl_replicas.common.policy import Policy
from rl_replicas.common.value_function import ValueFunction
from rl_replicas import log

logger = log.get_logger(__name__)

class PPO(OnPolicyAlgorithm):
  """
  Proximal Policy Optimization (by clipping) with early stopping based on approximate KL

  :param policy: (Policy) The policy
  :param value_function: (ValueFunction) The value function
  :param env: (gym.Env or str) The environment to learn from
  :param clip_range: (float) The limit on the likelihood ratio between policies for clipping in the policy objective.
  :param max_kl_divergence: (float) The limit on the KL divergence between policies for early stopping.
  :param gamma: (float) Discount factor
  :param gae_lambda: (float) Factor for trade-off of bias vs variance for Generalized Advantage Estimator. Equivalent to classic advantage when set to 1.
  :param seed: (int) The seed for the pseudo-random generators
  :param n_value_gradients (int): Number of gradient descent steps to take on value function per epoch.
  """
  def __init__(
    self,
    policy: Policy,
    value_function: ValueFunction,
    env: gym.Env,
    clip_range: float = 0.2,
    max_kl_divergence: float = 0.01,
    gamma: float = 0.99,
    gae_lambda: float = 0.97,
    seed: Optional[int] = None,
    n_policy_gradients: int = 80,
    n_value_gradients: int = 80,
  ) -> None:
    super().__init__(
      policy=policy,
      value_function=value_function,
      env=env,
      gamma=gamma,
      gae_lambda=gae_lambda,
      seed=seed,
      n_value_gradients=n_value_gradients
    )

    self.clip_range = clip_range
    self.n_policy_gradients = n_policy_gradients
    self.max_kl_divergence = max_kl_divergence

    self.old_policy: Policy = copy.deepcopy(self.policy)

  def train(
    self,
    one_epoch_experience: OneEpochExperience
  ) -> None:
    observations: torch.Tensor = one_epoch_experience['observations']
    actions: torch.Tensor = one_epoch_experience['actions']
    advantages: torch.Tensor = one_epoch_experience['advantages']

    # Normalize advantage
    advantages = (advantages - torch.mean(advantages)) / torch.std(advantages)

    # for logging
    with torch.no_grad():
      policy_dist: Categorical = self.policy(observations)
      policy_loss_before: torch.Tensor = self.compute_policy_loss(
                                           observations,
                                           actions,
                                           advantages
                                          ).detach()
    log_probs: torch.Tensor = policy_dist.log_prob(actions)
    entropies: torch.Tensor = policy_dist.entropy()

    # Train policy
    for i in range(self.n_policy_gradients):
      policy_loss: torch.Tensor = self.compute_policy_loss(observations, actions, advantages)
      approximate_kl_divergence: torch.Tensor = self.compute_approximate_kl_divergence(
                                                  observations,
                                                  actions
                                                ).detach()
      if approximate_kl_divergence > 1.5 * self.max_kl_divergence:
        logger.info('Early stopping at update {} due to reaching max KL divergence.'.format(i))
        break

      self.policy.optimizer.zero_grad()
      policy_loss.backward()
      self.policy.optimizer.step()

    self.old_policy.load_state_dict(self.policy.state_dict())

    discounted_returns: torch.Tensor = one_epoch_experience['discounted_returns']

    # for logging
    with torch.no_grad():
      value_loss_before: torch.Tensor = self.compute_value_loss(observations, discounted_returns)

    # Train value function
    for _ in range(self.n_value_gradients):
      value_loss: torch.Tensor = self.compute_value_loss(observations, discounted_returns)
      self.value_function.optimizer.zero_grad()
      value_loss.backward()
      self.value_function.optimizer.step()

    logger.info('Policy Loss:            {:<8.3g}'.format(policy_loss_before))
    logger.info('Avarage Entropy:        {:<8.3g}'.format(torch.mean(entropies)))
    logger.info('Log Prob STD:           {:<8.3g}'.format(torch.std(log_probs)))
    logger.info('KL divergence:          {:<8.3g}'.format(approximate_kl_divergence))

    logger.info('Value Function Loss:    {:<8.3g}'.format(value_loss_before))

    if self.writer:
      self.writer.add_scalar(
        'policy/loss',
        policy_loss_before,
        self.current_total_steps
      )
      self.writer.add_scalar(
        'policy/avarage_entropy',
        torch.mean(entropies),
        self.current_total_steps
      )
      self.writer.add_scalar(
        'policy/log_prob_std',
        torch.std(log_probs),
        self.current_total_steps
      )
      self.writer.add_scalar(
        'policy/approximate_kl_divergence',
        approximate_kl_divergence,
        self.current_total_steps
      )

      self.writer.add_scalar(
        'value/loss',
        value_loss_before,
        self.current_total_steps
      )

  def compute_policy_loss(
    self,
    observations: torch.Tensor,
    actions: torch.Tensor,
    advantages: torch.Tensor
  ) -> torch.Tensor:
    policy_dist: Categorical = self.policy(observations)
    log_probs: torch.Tensor = policy_dist.log_prob(actions)

    with torch.no_grad():
      old_policy_dist: Categorical = self.old_policy(observations)
      old_log_probs: torch.Tensor = old_policy_dist.log_prob(actions)

    # Calculate surrogate
    likelihood_ratio: torch.Tensor = torch.exp(log_probs - old_log_probs)
    surrogate: torch.Tensor = likelihood_ratio * advantages

    # Clipping the constraint
    likelihood_ratio_clip: torch.Tensor = torch.clamp(
                                            likelihood_ratio,
                                            min=1 - self.clip_range,
                                            max=1 + self.clip_range
                                          )

    # Calculate surrotate clip
    surrogate_clip: torch.Tensor = likelihood_ratio_clip * advantages

    policy_loss: torch.Tensor = -torch.min(surrogate, surrogate_clip).mean()

    return policy_loss

  def compute_approximate_kl_divergence(
    self,
    observations: torch.Tensor,
    actions: torch.Tensor
  ) -> torch.Tensor:
    with torch.no_grad():
      policy_dist: Categorical = self.policy(observations)
      log_probs: torch.Tensor = policy_dist.log_prob(actions)

      old_policy_dist: Categorical = self.old_policy(observations)
      old_log_probs: torch.Tensor = old_policy_dist.log_prob(actions)

    approximate_kl_divergence: torch.Tensor = old_log_probs - log_probs

    return torch.mean(approximate_kl_divergence)

  def compute_value_loss(
    self,
    observations: torch.Tensor,
    discounted_returns: torch.Tensor
  ) -> torch.Tensor:
    values: torch.Tensor = self.value_function(observations)
    squeezed_values: torch.Tensor = torch.squeeze(values, -1)
    value_loss: torch.Tensor = F.mse_loss(squeezed_values, discounted_returns)

    return value_loss



from setuptools import setup, find_packages

VERSION = '0.0.1'

long_description = """

# Quick reference

* **Github Repository**: [https://github.com/yamatokataoka/reinforcement-learning-replications](https://github.com/yamatokataoka/reinforcement-learning-replications)

# What is Reinforcement Learning Replications?

Reinforcement Learning Replications is a set of Pytorch implementations of reinforcement learning algorithms.

"""

setup(
  name='rl_replicas',
  version=VERSION,
  description='Reinforcement Learning Replications is a set of Pytorch implementations of reinforcement learning algorithms.',
  long_description=long_description,
  long_description_content_type='text/markdown',
  author='Yamato Kataoka',
  url='https://github.com/yamatokataoka/reinforcement-learning-replications',
  packages=find_packages(),
  license='MIT',
  keywords='reinforcement-learning',
  install_requires=[
    'gym[atari, box2d, classic_control]',
    'numpy',
    'scipy',
    'torch>=1.4.0',
    'tensorboard'
  ],
  python_requires='>=3.6'
)



from typing import Optional

import torch
from torch.nn import functional as F
from torch.distributions.categorical import Categorical
import gym

from rl_replicas.common.base_algorithms.on_policy_algorithm import OnPolicyAlgorithm, OneEpochExperience
from rl_replicas.common.policy import Policy
from rl_replicas.common.value_function import ValueFunction
from rl_replicas import log

logger = log.get_logger(__name__)

class VPG(OnPolicyAlgorithm):
  """
  Vanilla Policy Gradient (REINFORCE) with GAE for advantage estimation

  VPG, also known as Reinforce, trains stochastic policy in an on-policy way.

  :param policy: (Policy) The policy
  :param value_function: (ValueFunction) The value function
  :param env: (gym.Env or str) The environment to learn from
  :param gamma: (float) Discount factor
  :param gae_lambda: (float) Factor for trade-off of bias vs variance for Generalized Advantage Estimator. Equivalent to classic advantage when set to 1.
  :param seed: (int) The seed for the pseudo-random generators
  :param n_value_gradients (int): Number of gradient descent steps to take on value function per epoch.
  """
  def __init__(
    self,
    policy: Policy,
    value_function: ValueFunction,
    env: gym.Env,
    gamma: float = 0.99,
    gae_lambda: float = 0.97,
    seed: Optional[int] = None,
    n_value_gradients: int = 80
  ) -> None:
    super().__init__(
      policy=policy,
      value_function=value_function,
      env=env,
      gamma=gamma,
      gae_lambda=gae_lambda,
      seed=seed,
      n_value_gradients=n_value_gradients
    )

  def train(
    self,
    one_epoch_experience: OneEpochExperience
  ) -> None:
    observations: torch.Tensor = one_epoch_experience['observations']
    actions: torch.Tensor = one_epoch_experience['actions']
    advantages: torch.Tensor = one_epoch_experience['advantages']

    # Normalize advantage
    advantages = (advantages - torch.mean(advantages)) / torch.std(advantages)

    policy_dist: Categorical = self.policy(observations)
    log_probs: torch.Tensor = policy_dist.log_prob(actions)

    policy_loss: torch.Tensor = -torch.mean(log_probs * advantages)

    # for logging
    policy_loss_before: torch.Tensor = policy_loss.detach()
    entropies: torch.Tensor = policy_dist.entropy().detach()

    # Train policy
    self.policy.optimizer.zero_grad()
    policy_loss.backward()
    self.policy.optimizer.step()

    discounted_returns: torch.Tensor = one_epoch_experience['discounted_returns']

    # for logging
    with torch.no_grad():
      value_loss_before: torch.Tensor = self.compute_value_loss(observations, discounted_returns)

    # Train value function
    for _ in range(self.n_value_gradients):
      value_loss: torch.Tensor = self.compute_value_loss(observations, discounted_returns)
      self.value_function.optimizer.zero_grad()
      value_loss.backward()
      self.value_function.optimizer.step()

    logger.info('Policy Loss:            {:<8.3g}'.format(policy_loss_before))
    logger.info('Avarage Entropy:        {:<8.3g}'.format(torch.mean(entropies)))
    logger.info('Log Prob STD:           {:<8.3g}'.format(torch.std(log_probs)))

    logger.info('Value Function Loss:    {:<8.3g}'.format(value_loss_before))

    if self.writer:
      self.writer.add_scalar(
        'policy/loss',
        policy_loss_before,
        self.current_total_steps
      )
      self.writer.add_scalar(
        'policy/avarage_entropy',
        torch.mean(entropies),
        self.current_total_steps
      )
      self.writer.add_scalar(
        'policy/log_prob_std',
        torch.std(log_probs),
        self.current_total_steps
      )

      self.writer.add_scalar(
        'value/loss',
        value_loss_before,
        self.current_total_steps
      )

  def compute_value_loss(
    self,
    observations: torch.Tensor,
    discounted_returns: torch.Tensor
  ) -> torch.Tensor:
    values: torch.Tensor = self.value_function(observations)
    squeezed_values: torch.Tensor = torch.squeeze(values, -1)
    value_loss: torch.Tensor = F.mse_loss(squeezed_values, discounted_returns)

    return value_loss



import random
from typing import List, Tuple

import numpy as np
import scipy.signal
import torch

def discount_cumulative_sum(vector: np.ndarray, discount: float) -> np.ndarray:
  """
  Compute discounted cumulative sums of vector.

  :param vector: (np.ndarray) An target vector
  e.g. [x0,
        x1,
        x2]
  :param discount: (float) Discount factor
  :return: discounted cumulative sums of an vector
  e.g. [x0 + discount * x1 + discount^2 * x2,
        x1 + discount * x2,
        x2]
  """
  return scipy.signal.lfilter([1], [1, -discount], vector[::-1], axis=0)[::-1]

def seed_random_generators(seed: int) -> None:
    """
    Set the seed of the pseudo-random generators
    (python, numpy, pytorch)

    :param seed: (int)
    """
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)

def gae(
  rewards: np.ndarray,
  gamma: float,
  values: np.ndarray,
  gae_lambda: float
) -> np.ndarray:
  """
  Compute Generalized Advantage Estimation (GAE)

  :param rewards: (np.ndarray) Rewards for all states
  :param gamma: (float) Discount factor
  :param values: (np.ndarray) Values for all states
  :param gae_lambda: (float) A smoothing parameter for reducing the variance
  :return gaes: GAEs for all states
  """
  deltas: np.ndarray = rewards[:-1] + gamma * values[1:] - values[:-1]
  gaes: np.ndarray = discount_cumulative_sum(deltas, gamma * gae_lambda)

  return gaes

def unflatten_tensors(flattened: np.ndarray, tensor_shapes: List[torch.Size]) -> List[torch.Tensor]:
    """
    Unflatten a flattened tensors into a list of tensors.

    :param flattened: (np.ndarray) Flattened tensors.
    :param tensor_shapes: (List[torch.Size]) Tensor shapes.
    :return: (List[np.ndarray]) Unflattened list of tensors.
    """
    tensor_sizes = list(map(np.prod, tensor_shapes))
    indices = np.cumsum(tensor_sizes)[:-1]

    return [np.reshape(pair[0], pair[1]) for pair in zip(np.split(flattened, indices), tensor_shapes)]
